#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <getopt.h>

struct passwd *pw, *pw_euid;
struct group *gr, *gr2, *gr_egid;
uid_t uid, euid;
gid_t gid, egid;


struct passwd *
xgetpwuid(int u)
{
	struct passwd *p = getpwuid(u);
	
	if (p == NULL) {
		printf("Record not found or an error.\n");
		exit(1);
	}
	
	return p;
}

struct group *
xgetgrgid(int gd)
{
	struct group *g = getgrgid(gd);
	
	if (g == NULL) {
		printf("Record not found or an error.\n");
		exit(1);
	}
	
	return g;
}

struct passwd *
xgetpwnam(char *pwnam)
{
	struct passwd *p = getpwnam(pwnam);
	
	if (p == NULL) {
		printf("Record not found or an error.\n");
		exit(1);
	}
	
	return p;
}

struct group *
xgetgrnam(char *grnam)
{
	struct group *g = getgrnam(grnam);
	
	if (g == NULL) {
		printf("Record not found or an error.\n");
		exit(1);
	}
	
	return g;
}

void no_opt(void)
{
	if ((uid == euid) && (gid == egid))
		printf("uid=%u(%s) gid=%u(%s) ", 
		       uid, pw->pw_name, gid, gr2->gr_name);
	else if ((uid == euid) && (gid != egid))		
		printf("uid=%u(%s) gid=%u(%s) egid=%u(%s) ", 
		       uid, pw->pw_name, gid, gr2->gr_name, egid, gr_egid->gr_name);
	else if ((uid != euid) && (gid == egid))		
		printf("uid=%u(%s) gid=%u(%s) euid=%u(%s) ", 
		       uid, pw->pw_name, gid, gr2->gr_name, euid, pw_euid->pw_name);
	else 
		printf("uid=%u(%s) gid=%u(%s) euid=%u(%s) egid=%u(%s) ", 
		       uid, pw->pw_name, gid, gr2->gr_name, euid, pw_euid->pw_name, egid, gr_egid->gr_name);
	
	printf("groups=%u(%s)", gid, gr2->gr_name); 	
	while(1) {
		gr = getgrent();
		
		if (gr == NULL) break;

		int i = 0;
		while(gr->gr_mem[i] != NULL) {
			if (strcmp(pw->pw_name, gr->gr_mem[i]) == 0)
				printf(",%u(%s)",gr->gr_gid, gr->gr_name);
			i++;
		}
	}
	printf("\n");
	endgrent();
}

void opt_G(int q)
{
	if (q == 0) {
		printf("%u", gid);
		
		while(1) {
			gr = getgrent();
						
			if (gr == NULL) break;
			
			int i = 0;
			while(gr->gr_mem[i] != NULL) {
				if (strcmp(pw->pw_name, gr->gr_mem[i]) == 0)
					printf(" %u",gr->gr_gid);
				
				i++;
			}
		}
	} else {
		printf("%s", gr2->gr_name);
		
		while(1) {
			gr = getgrent();
						
			if (gr == NULL) break;
			
			int i = 0;
			while(gr->gr_mem[i] != NULL) {
				if (strcmp(pw->pw_name, gr->gr_mem[i]) == 0)
					printf(" %s",gr->gr_name);
				
				i++;
			}
		}
	}
	endgrent();
	printf("\n");	
}

void 
opt_g(int q)
{
	if (q == 0)
		printf("%u\n", egid);
	else if (q == 1)
		printf("%s\n", gr_egid->gr_name);
	else if (q == 2)
		printf("%u\n", gid);
	else
		printf("%s\n", gr2->gr_name);
}

void 
opt_u(int q)
{
	if (q == 0)
		printf("%u\n", euid);
	else if (q == 1)
		printf("%s\n", pw_euid->pw_name);
	else if (q == 2)
		printf("%u\n", uid);
	else
		printf("%s\n", pw->pw_name);
}

int is_g = 0;
int is_G = 0;
int is_u = 0;
int is_r = 0;
int is_n = 0;
int non_opt = 0;

void
get_opt(int argc, char *argv[], char *string)
{
	char opt;
	opterr = 0;
	
	while(1) {
		opt = getopt(argc, argv, string);
		if (opt == -1)
			break;
		
		switch(opt) {
			case 'G':
				is_G = 1;
				break;
			case 'g':
				is_g = 1;
				break;
			case 'n':
				is_n = 1;
				break;
			case 'r':
				is_r = 1;
				break;
			case 'u':
				is_u = 1;
				break;
			case '?':
				break;
		}
	}

	if (optind == 1)
		non_opt = 1;
}

void 
opt_tok()
{
	if (non_opt) {
		no_opt();
	} else if (is_g && is_G) {
		printf("Invalid format options\n");
	} else if (is_g && is_u) {
		printf("Invalid format options\n");
	} else if (is_g && is_u) {
		printf("Invalid format options\n");
	} else if (is_G && is_r) {
		printf("Invalid format options\n"); 
	} else if (is_G && is_n) {
		opt_G(1);
	} else if (is_G) {
		opt_G(0);
	} else if (is_g && is_n && is_r) {
		opt_g(3);
	} else if (is_g && is_n) {
		opt_g(1);
	} else if (is_g && is_r) {
		opt_g(2);
	} else if (is_g) {
		opt_g(0);
	} else if (is_u && is_n && is_r) {
		opt_u(3);
	} else if (is_u && is_n) {
		opt_u(1);
	} else if (is_u && is_r) {
		opt_u(2);
	} else if (is_u) {
		opt_u(0);
	} else 
		printf("Invalid format options\n");		
}

int main(int argc, char *argv[])
{
	get_opt(argc, argv, "gGrnu");

	if (argv[optind] == NULL) {
		uid = getuid();
		gid = getgid();
		pw = xgetpwuid(uid);
		euid = geteuid();
		egid = getegid();
		gr2 = xgetgrgid(gid);
		pw_euid = xgetpwuid(euid);
		gr_egid = xgetgrgid(egid);
	} else {
		pw = xgetpwnam(argv[optind]);
		gr2 = xgetgrnam(argv[optind]);
		uid = pw->pw_uid;
		gid = gr2->gr_gid;
		euid = geteuid();
		egid = getegid();
		pw_euid = xgetpwuid(euid);
		gr_egid = xgetgrgid(egid);
	}
	opt_tok();
  return EXIT_SUCCESS;
}